//    gcode "remaining material" previewer
//    Copyright (C) 2007 Jeff Epler <jepler@unpythonic.net>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <Python.h>
#include <numarray.h>
#include <libnumarray.h>
#include <math.h>
#include <GL/gl.h>
#include <stdio.h>

#define d(x,y) D[x*dh+y]
#define p(x,y) P[x*pw+y]

static float min(float a, float b) { if(a < b) return a; return b; }
static float max(float a, float b) { if(a > b) return a; return b; }

static void
put1(int x0, int y0, float z, int dw, int dh, float *D, int pw, float *P) {
    int x, y;
    for(x = 0; x < pw; x++) {
	int xd = x + x0;
	if(xd < 0) continue;
	if(xd >= dw) break;

	for(y = 0; y < pw; y++) {
	    int yd = y + y0;
	    if(yd < 0) continue;
	    if(yd >= dh) break;

	    d(xd,yd) = min(d(xd,yd), p(x,y) + z);
	}
    }
}

static void
putline(int x0, int y0, float z0, int x1, int y1, float z1,
	int dw, int dh, float *D, int pw, float *P) {
    double dx, dy, dz;
    int t1, t;

    if(x0 == x1 && y0 == y1) {
	put1(x0, y0, min(z0, z1), dw, dh, D, pw, P);
	return;
    }
    dx = x1-x0;
    dy = y1-y0;
    dz = z1-z0;
    t1 = max(fabs(dx), fabs(dy));

    dx = dx / t1;
    dy = dy / t1;
    dz = dz / t1;

    for(t=0; t <= t1; t ++) {
	put1(x0 + t*dx, y0 + t*dy, z0 + t*dz, dw, dh, D, pw, P);
    }
}

PyObject *py_putline(PyObject *s, PyObject *o) {
    int x0, y0, x1, y1;
    double z0, z1;
    PyObject *Do, *Po;
    PyArrayObject *Da, *Pa;

    if(!PyArg_ParseTuple(o, "iidiidOO", &x0, &y0, &z0, &x1, &y1, &z1, &Do, &Po))
	return 0;

    Da = NA_InputArray(Do, tFloat32, NUM_C_ARRAY);
    Pa = NA_InputArray(Po, tFloat32, NUM_C_ARRAY);

    putline(x0, y0, z0, x1, y1, z1, Da->dimensions[0], Da->dimensions[1],
	    NA_OFFSETDATA(Da), Pa->dimensions[0], NA_OFFSETDATA(Pa));

    Py_RETURN_NONE;
}

static inline void
triangle(float *D, int dw, int dh, double resp, int x1, int y1,
	int x2, int y2, int x3, int y3) {
    double dx1, dy1, dx2, dy2, dz1, dz2;
    double x1_, y1_, x2_, y2_, x3_, y3_;
    double nx, ny, nz;
    double z1, z2, z3;

    z1 = d(x1,y1);
    z2 = d(x2,y2);
    z3 = d(x3,y3);

    x1_ = x1 * resp;
    x2_ = x2 * resp;
    x3_ = x3 * resp;
    y1_ = y1 * resp;
    y2_ = y2 * resp;
    y3_ = y3 * resp;

    dx1 = x2_-x1_;
    dx2 = x3_-x1_;
    dy1 = y2_-y1_;
    dy2 = y3_-y1_;
    dz1 = z2-z1;
    dz2 = z3-z1;


    nx = dy1 * dz2 - dz1 * dy2;
    ny = dz1 * dx2 - dx1 * dz2;
    nz = dx1 * dy2 - dy1 * dx2;

    glNormal3f(nx, ny, nz);
    glVertex3f(x3_, y3_, z3);
    glVertex3f(x2_, y2_, z2);
    glVertex3f(x1_, y1_, z1);
}
PyObject *py_triangles(PyObject *s, PyObject *o) {
    int x, y;
    PyObject *Do;
    PyArrayObject *Da;
    int dw, dh;
    double res, resp;
    float *D;

    if(!PyArg_ParseTuple(o, "dO", &res, &Do)) return 0;

    Da = NA_InputArray(Do, tFloat32, NUM_C_ARRAY);
    dw = Da->dimensions[0];
    dh = Da->dimensions[1];
    D = NA_OFFSETDATA(Da);

    resp = 1.0/res;

    glBegin(GL_TRIANGLES);
    for(x=0; x < dw-1; x++) {
	for(y=0; y < dh-1; y++) {
	    triangle(D, dw, dh, resp, x,y, x,y+1, x+1,y);
	    triangle(D, dw, dh, resp, x,y+1, x+1,y+1, x+1,y);
	}
    }
    glEnd();
    Py_RETURN_NONE;
}

static PyMethodDef methods[] = {
    {"putline", py_putline, METH_VARARGS, "putline(x0,y0,z0, x1,y1,z1, D, P)\n"
	"Draw a line from 0..1 on D with tool P"},
    {"triangles", py_triangles, METH_VARARGS, "triangles(res, P)\n"
	"Render the triangles for mesh P of resolution res in OpenGL"},
    {NULL}
};

void init_gdepth(void) {
    import_libnumarray();
    Py_InitModule("_gdepth", methods);
}
